#!/usr/bin/env python
"""
    Calibrate the Telescope from the RESTful API

    Copyright (c) Tim Molteno 2017-2022.

    This tool uses  high-dimensional optimisation to calculate the gains and phases of the 24 antennas
    of the telescope.
"""
import matplotlib

matplotlib.use("agg")
import matplotlib.pyplot as plt

import argparse
import numpy as np
import time

from copy import deepcopy

import itertools

from tart.operation import settings
from tart.imaging import visibility
from tart.imaging import calibration
from tart.imaging import synthesis
from tart.imaging import elaz

from tart.util.angle import from_rad

from tart_tools import api_imaging
from tart_tools import api_handler

triplets = None
ij_index = None
jk_index = None
ik_index = None


def split_param(x):
    rot_degrees = x[0]
    re = np.concatenate(([1], x[1:24]))
    im = np.concatenate(([0], x[24:47]))
    gains = np.sqrt(re * re + im * im)
    phase_offsets = np.arctan2(im, re)
    return rot_degrees, gains, phase_offsets


def join_param(rot_degrees, gains, phase_offsets):
    ret = np.zeros(47)
    ret[0] = rot_degrees
    z = gains[1:24] * np.exp(phase_offsets[1:24] * 1j)
    ret[1:24] = z.real
    ret[24:47] = z.imag
    return ret


def param_to_json(x):
    rot_degrees, gains, phase_offsets = split_param(x)
    ret = {
        "gain": np.round(gains, 4).tolist(),
        "rot_degrees": rot_degrees,
        "phase_offset": np.round(phase_offsets, 4).tolist(),
    }
    return ret


def output_param(x, fp=None):
    ret = param_to_json(x)
    print(ret)
    if fp is None:
        print(json.dumps(ret, indent=4, separators=(",", ": ")))
    else:
        json.dump(ret, fp, indent=4, separators=(",", ": "))


def calc_score_aux(opt_parameters, measurements, window_deg, original_positions):
    global triplets, ij_index, jk_index, ik_index, masks
    rot_degrees, gains, phase_offsets = split_param(opt_parameters)

    ret_zone = 0.0
    ret_std = 0.0

    ant_idxs = np.arange(24)

    for i, m in enumerate(measurements):
        cv, ts, src_list = m

        cv.set_phase_offset(ant_idxs, phase_offsets)
        cv.set_gain(ant_idxs, gains)
        api_imaging.rotate_vis(rot_degrees, cv, original_positions)

        n_bin = 2 ** 7
        cal_ift, cal_extent, n_fft, bin_width = api_imaging.image_from_calibrated_vis(
            cv, nw=n_bin / 4, num_bin=n_bin
        )

        abs_ift = np.abs(cal_ift)
        ift_std = np.std(abs_ift)
        ift_scaled = abs_ift / ift_std

        ret_std += -np.sqrt(ift_scaled.max())  # Peak signal to noise.

        if masks[i] is None:
            print("Creating mask")
            mask = np.zeros_like(ift_scaled)

            for s in src_list:
                x0, y0 = s.get_px(n_fft)
                d = 2*s.deg_to_pix(n_fft, window_deg)
                for y in range(mask.shape[0]):
                    for x in range(mask.shape[1]):
                        r2 = (y - y0)**2 + (x - x0)**2
                        p = np.exp(-r2/d)
                        mask[y, x] += p
            masks[i] = mask / np.sum(mask)

        mask = masks[i]
        zone_score = -np.sum(mask * ift_scaled)
        #zones = []
        #for s in src_list:
            ## Get a pixel window around each source...
            #x_min, x_max, y_min, y_max, area = s.get_px_window(n_fft, window_deg)
            ## Integrate the brightness in a window around that source
            #s_px = ift_scaled[y_min:y_max, x_min:x_max]
            #zones.append(np.sum(s_px) / area)
        #zones = np.array(zones)
        #zone_score = -np.mean(zones)

        ret_zone += 4 * zone_score

    if N_IT % 100 == 0:
        print(f"S/N {ret_std:04.2f}, ZONE: {ret_zone:04.2f}")

    return (
        (ret_std + ret_zone) / len(measurements),
        ift_scaled,
        src_list,
        n_fft,
        bin_width,
        mask,
    )


def load_data_from_json(
    vis_json, src_json, config, gains, phases, flag_list, el_threshold
):

    cv, ts = api_imaging.vis_calibrated(vis_json, config, gains, phases, flag_list)
    src_list = elaz.from_json(src_json, el_threshold)
    return cv, ts, src_list


def get_window(x_i, y_i, d):
    x_min = int(np.floor(x_i - d))
    x_max = int(np.ceil(x_i + d))
    y_min = int(np.floor(y_i - d))
    y_max = int(np.ceil(y_i + d))
    return x_min, x_max, y_min, y_max


def calc_score(
    opt_parameters,
    config,
    measurements,
    window_deg,
    original_positions,
    update=False,
    show=False,
):
    global N_IT, method, output_directory, f_vs_iteration

    ret, ift_scaled, src_list, n_fft, bin_width, mask = calc_score_aux(
        opt_parameters, measurements, window_deg, original_positions
    )

    if N_IT % 100 == 0:
        print(f"Iteration {N_IT}, score={ret:04.2f}")
        f_vs_iteration.append(ret)

    if N_IT % 1000 == 0:

        #ift_sel = np.zeros_like(ift_scaled)

        #for s in src_list:
            #x_min, x_max, y_min, y_max, area = s.get_px_window(n_fft, window_deg)
            #ift_sel[y_min:y_max, x_min:x_max] = ift_scaled[y_min:y_max, x_min:x_max]

        ift_sel = ift_scaled*mask
        x_list, y_list = elaz.get_source_coordinates(src_list)

        plt.figure()
        plt.imshow(
            ift_sel,
            extent=[-1, 1, -1, 1],
            vmin=0,
        )  # vmax=8
        plt.colorbar()
        plt.xlim(1, -1)
        plt.ylim(-1, 1)
        plt.scatter(x_list, y_list, c="red", s=5)
        plt.xlabel("East-West")
        plt.ylabel("North-South")
        plt.tight_layout()
        plt.savefig("{}/opt_slice_{:05d}.png".format(output_directory, N_IT))
        plt.close()

        plt.figure()
        plt.imshow(
            ift_scaled,
            extent=[-1, 1, -1, 1],
            vmin=0,
        )  # vmax=8
        plt.colorbar()
        plt.xlim(1, -1)
        plt.ylim(-1, 1)
        plt.title(ret)
        plt.scatter(x_list, y_list, c="red", s=5)
        plt.xlabel("East-West")
        plt.ylabel("North-South")
        plt.tight_layout()
        plt.savefig(
            "{}/{}_{:5.3f}_opt_full_{:05d}.png".format(
                output_directory, method, ret, N_IT
            )
        )
        plt.close()
    N_IT += 1
    return ret


from scipy import optimize
import json


class MyTakeStep(object):
    def __init__(self, stepsize=0.5):
        self.stepsize = stepsize

    def __call__(self, x):
        s = self.stepsize
        x[0] += np.random.uniform(-2.0 * s, 2.0 * s)
        x[1:] += np.random.uniform(-s, s, x[1:].shape)
        return x


def bh_callback(x, f, accepted):
    global output_directory, bh_basin_progress, N_IT
    print("BH f={} accepted {}".format(f, int(accepted)))
    output_param(x)
    if accepted:
        bh_basin_progress.append([N_IT, float(f)])
        with open("{}/bh_basin_progress.json".format(output_directory), "w") as fp:
            json.dump(bh_basin_progress, fp, indent=4, separators=(",", ": "))

        with open(
            "{}/BH_basin_{:5.3f}_{}.json".format(output_directory, float(f), N_IT), "w"
        ) as fp:
            output_param(x, fp)


def de_callback(xk, convergence):
    print("DE at {} conv={}".format(xk, convergence))
    output_param(xk)


def main():
    parser = argparse.ArgumentParser(
        description="Calibrate the tart telescope from API.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--api",
        required=False,
        default="https://tart.elec.ac.nz/signal",
        help="Telescope API server URL.",
    )
    parser.add_argument(
        "--file",
        required=False,
        default="calibration_data.json",
        help="Calibration Output JSON file.",
    )

    #parser.add_argument(
        #"--influx",
        #required=False,
        #default=None,
        #help="Use an influxdb  (https://tartinflux.max.ac.nz/) for cal data.",
    #)

    parser.add_argument("--show", action="store_true", help="show instead of save.")
    parser.add_argument(
        "--cold-start", action="store_true", help="Start from zero knowledge of gains."
    )
    parser.add_argument(
        "--get-gains",
        action="store_true",
        help="Start from current knowledge of gains.",
    )
    parser.add_argument("--dir", required=False, default=".", help="Output directory.")
    parser.add_argument(
        "--method",
        required=False,
        default="LB",
        help="Optimization Method [NM, LB, DE, BH]",
    )
    parser.add_argument(
        "--iterations",
        required=False,
        type=int,
        default=300,
        help="Number of iterations for basinhopping",
    )
    parser.add_argument(
        "--elevation", type=float, default=30.0, help="Elevation threshold for sources]"
    )

    parser.add_argument(
        '--ignore', nargs='+', type=int, help="Specify the list of antennas to zero out.")


    ARGS = parser.parse_args()

    # Load calibration data
    with open(ARGS.file, "r") as json_file:
        calib_info = json.load(json_file)

    info = calib_info["info"]
    ant_pos = calib_info["ant_pos"]
    config = settings.from_api_json(info["info"], ant_pos)

    flag_list = []  # [4, 5, 14, 22] # optionally remove some unused antennas

    method = ARGS.method
    output_directory = ARGS.dir
    f_vs_iteration = []

    original_positions = deepcopy(config.get_antenna_positions())

    gains_json = calib_info["gains"]
    print(gains_json["gain"])
    gains = np.asarray(gains_json["gain"])
    phase_offsets = np.asarray(gains_json["phase_offset"])
    print(gains)
    print(phase_offsets)

    if ARGS.cold_start and ARGS.get_gains:
        raise Exception("ERROR: Cannot Have both cold_start and get-gains specified")

    if ARGS.cold_start:
        gains = np.ones(len(gains_json["gain"]))
        phase_offsets = np.zeros(len(gains_json["phase_offset"]))
    if ARGS.get_gains:
        api = api_handler.APIhandler(ARGS.api)
        gains_json = api.get("calibration/gain")
        gains = np.asarray(gains_json["gain"])
        phase_offsets = np.asarray(gains_json["phase_offset"])

    config = settings.from_api_json(info["info"], ant_pos)

    init_parameters = join_param(0.0, gains, phase_offsets)
    output_param(init_parameters)

    masks = []
    measurements = []
    for d in calib_info["data"]:
        vis_json, src_json = d
        cv, ts, src_list = load_data_from_json(
            vis_json,
            src_json,
            config,
            gains,
            phase_offsets,
            flag_list,
            el_threshold=ARGS.elevation,
        )
        measurements.append([cv, ts, src_list])
        masks.append(None)

    N_IT = 0
    window_deg = 5.0

    s = calc_score(
        init_parameters,
        config,
        measurements,
        window_deg,
        original_positions,
        update=False,
        show=False,
    )

    f = lambda param: calc_score(
        param,
        config,
        measurements,
        window_deg,
        original_positions,
        update=False,
        show=False,
    )

    bounds = [(-5, 5)]  # Bounds for the rotation parameter
    for i in range(46):
        bounds.append((-1.2, 1.2))
        # Bounds for all other parameters (real and imaginary components)

    zero_list = ARGS.ignore
    if zero_list is not None:
        print(f"Ignoring antennas {zero_list}")
        for i in range(24):
            if i in zero_list:
                bounds[i] = (-0.01, 0.01)
                bounds[i+23] = (-0.01, 0.01)

    np.random.seed(555)  # Seeded to allow replication.

    if method == "NM":
        ret = optimize.minimize(f, init_parameters, method="Nelder-Mead", tol=1e-5)
    if method == "LB":
        ret = optimize.minimize(f, init_parameters, method="L-BFGS-B", bounds=bounds)
    if method == "DE":
        ret = optimize.differential_evolution(f, bounds, disp=True)
    if method == "BH":
        bh_basin_progress = [[0, s]]
        minimizer_kwargs = {
            "method": "L-BFGS-B",
            "jac": False,
            "bounds": bounds,
            "tol": 1e-5,
            "options": {"maxcor": 48},
        }
        ret = optimize.basinhopping(
            f,
            init_parameters,
            niter=ARGS.iterations,
            T=0.5,
            stepsize=2.0,
            disp=True,
            minimizer_kwargs=minimizer_kwargs,
            callback=bh_callback,
        )
        with open("{}/bh_basin_progress.json".format(output_directory), "w") as fp:
            json.dump(bh_basin_progress, fp, indent=4, separators=(",", ": "))

    rot_degrees = ret.x[0]
    output_json = param_to_json(ret.x)
    output_json["message"] = ret.message
    output_json["optimum"] = ret.fun
    output_json["iterations"] = ret.nit

    new_positions = settings.rotate_location(
        rot_degrees, np.array(original_positions).T
    )
    print(new_positions)
    pos_list = (np.array(new_positions).T).tolist()
    print(pos_list)
    output_json["antenna_positions"] = pos_list

    with open("{}/{}_opt_json.json".format(output_directory, method), "w") as fp:
        json.dump(output_json, fp, indent=4, separators=(",", ": "))

    f_history_json = {}
    f_history_json["start"] = s
    f_history_json["history"] = f_vs_iteration

    with open("{}/{}_history.json".format(output_directory, method), "w") as fp:
        json.dump(f_history_json, fp, indent=4, separators=(",", ": "))



"""
    Image from the HTTP API

    Tim Molteno & Max Scheel 2017-2025.
"""
import logging
import os

import numpy as np
from tart.imaging import calibration, elaz, synthesis, visibility
from tart.imaging import imaging
from tart.util import utc

logger = logging.getLogger()


def vis_json_timestamp(vis_json):
    logger.info(f"Parsing {vis_json['timestamp']}")
    return utc.from_string(vis_json["timestamp"])


def vis_object_from_json(vis_json, config):
    ts = vis_json_timestamp(vis_json)
    ret = visibility.Visibility.from_config(config, ts)
    v_order = []
    bl_order = []
    for v in vis_json["data"]:
        v_order.append(complex(v["re"], v["im"]))
        bl_order.append([v["i"], v["j"]])
    ret.set_visibilities(v_order, bl_order)
    return ret


def vis_calibrated(vis_json, config, gains, phase_offset, flag_list=[]):
    vis = vis_object_from_json(vis_json, config)

    cal_vis = calibration.CalibratedVisibility(vis)
    for f in flag_list:
        cal_vis.flag_antenna(f)

    cal_vis.set_gain(np.arange(24), gains)
    cal_vis.set_phase_offset(np.arange(24), phase_offset)

    return cal_vis, vis.timestamp


def rotate_vis(rot_degrees, cv, reference_positions):
    return imaging.rotate_vis(rot_degrees, cv, reference_positions)


def image_from_calibrated_vis(cv, nw, num_bin):
    return imaging.image_from_calibrated_vis(cv, nw, num_bin)


def beam_from_calibrated_vis(cv, nw, num_bin):
    """
    Generate a beam (or Point Spread Function (PSF)
    for the antenna array
    """
    cal_syn = synthesis.Synthesis_Imaging([cv])

    return cal_syn.get_beam(nw=nw, num_bin=num_bin, use_kernel=False)


def get_uv_fits(cv):
    cal_syn = synthesis.Synthesis_Imaging([cv])
    return cal_syn.get_uvfits()


def make_image(plt, img, title, num_bins, source_json=None, healpix=False):
    if healpix:
        make_healpix_image(plt, img, title, num_bins, source_json)
    else:
        make_square_image(plt, img, title, num_bins, source_json)


def make_square_image(plt, img, title, num_bins, source_json=None):

    plt.figure(figsize=(8, 6), dpi=num_bins / 6)
    plt.title(title)

    print(f"Dynamic Range: {np.max(img)}")

    plt.imshow(img, extent=[-1, 1, -1, 1])

    plt.xlim(-1, 1)
    plt.ylim(-1, 1)
    cb = plt.colorbar()

    if source_json is not None:
        src_list = elaz.from_json(source_json, el_limit=20.0, jy_limit=1e4)
        output_list = []
        output_list.append(plt.Circle([0, 0], 1.0, color=(0.1, 0.1, 0.9), fill=False))
        for s in src_list:
            output_list.append(
                plt.Circle(s.get_lm(), 0.03, color=(0.9, 0.2, 0.3), fill=False)
            )
        ax = plt.gca()
        for circle in output_list:
            ax.add_artist(circle)

    # plt.grid()
    plt.xlabel("East-West")
    plt.ylabel("North-South")
    plt.tight_layout()


def make_healpix_image(plt, img, title, num_bins, source_json=None):
    """
    Writes out an image as a healpy image
    """
    import healpy as hp
    nside = hp.pixelfunc.get_min_valid_nside(num_bins * num_bins * 3 / 4)
    npix = hp.nside2npix(nside)

    pixels = np.arange(npix)
    m = np.zeros(npix) + hp.UNSEEN

    window_d = np.degrees(hp.nside2resol(nside))

    for i in pixels:
        theta, phi = hp.pix2ang(nside, i)
        if theta < np.pi / 2:
            el = np.degrees(np.pi / 2 - theta)
            az = np.degrees(phi)
            s = elaz.ElAz(el, az)
            x_min, x_max, y_min, y_max, area = s.get_px_window(
                num_bins, window_deg=window_d
            )
            s_px = img[y_min:y_max, x_min:x_max]

            m[i] = np.sum(s_px) / area

    hp.orthview(m, rot=(0, 90, 180), title=title, xsize=3000, cbar=False, half_sky=True)
    hp.graticule()
    if source_json is not None:
        src_list = elaz.from_json(source_json, el_limit=20.0, jy_limit=1e4)
        output_list = []
        for s in src_list:
            l, m = s.get_lm()
            output_list.append(
                plt.Circle([-l, m], 0.03, color=(0.9, 0.2, 0.3), fill=False)
            )
        ax = plt.gca()
        for circle in output_list:
            ax.add_artist(circle)

    # hp.projplot([float(sp.N(theta_actual)),], [float(sp.N(phi_actual)),], 'ro', rot=(0,90,0))
    # hp.projplot([float(sp.N(theta_actual2)),], [float(sp.N(phi_actual2)),], 'ro', rot=(0,90,0))


def save_fits_image(img, fname, timestamp, out_dir, header_dict={}):
    """
    This method saves a 2D array as a FITS image, and adds the minimum neccessary headers
    to get the image to be processable by MORESANE
    """
    import astropy.io.fits as pyfits

    hdu = pyfits.PrimaryHDU(img.astype(np.float32))
    hdulist = pyfits.HDUList([hdu])
    deg_per_pixel = 180.0 / len(img)
    prihdr = hdulist[0].header

    prihdr.set("CTYPE1", "RA---SIN")  # First parameter RA  ,  projection TANgential
    prihdr.set("CDELT1", deg_per_pixel)  # Degrees/pixel
    prihdr.set("CUNIT1", "deg")  # Degrees

    prihdr.set("CTYPE2", "DEC--SIN")  # Second parameter DEC,  projection TANgential
    prihdr.set("CDELT2", deg_per_pixel)  # Degrees/pixel
    prihdr.set("CUNIT2", "deg")  # Degrees

    prihdr.set("CROTA2", 0.00000)  # Rotation in degrees.

    prihdr.set("BMAJ", 0.0)  # Beam Major Axis (degrees)
    prihdr.set("BMIN", 90.0)  # Beam Minor Axis (degrees)
    prihdr.set("BPA", 0.0)  # Beam position angle

    prihdr.set("DATE", timestamp)
    prihdr.set("DATE-OBS", timestamp)

    prihdr.set("TIMESYS", "UTC")
    prihdr.set("INSTRUME", "TART")
    prihdr.set("TELESCOP", "TART")
    prihdr.set("OBSERVER", "CASA simulator")
    prihdr.set("ORIGIN", "tart_tools tart.elec.ac.nz ")

    prihdr.set("OBSRA", 2.889721000000e02)
    prihdr.set("OBSDEC", -7.466052777778e01)
    prihdr.set("OBSGEO-X", 5.111202828133e06)
    prihdr.set("OBSGEO-Y", 2.001309252764e06)
    prihdr.set("OBSGEO-Z", -3.237339358474e06)

    """
    SIMPLE  =                    T /Standard FITS
    BITPIX  =                  -32 /Floating point (32 bit)
    NAXIS   =                    4
    NAXIS1  =                 1024
    NAXIS2  =                 1024
    NAXIS3  =                    1
    NAXIS4  =                    1
    BSCALE  =   1.000000000000E+00 /PHYSICAL = PIXEL*BSCALE + BZERO
    BZERO   =   0.000000000000E+00
    BTYPE   = 'Intensity'
    OBJECT  = 'KAT7_1445_1x16_12h'

    BUNIT   = 'JY/BEAM '           /Brightness (pixel) unit
    EQUINOX =   2.000000000000E+03
    LONPOLE =   1.800000000000E+02
    LATPOLE =  -7.466052777778E+01
    PC001001=   1.000000000000E+00
    PC002001=   0.000000000000E+00
    PC003001=   0.000000000000E+00
    PC004001=   0.000000000000E+00
    PC001002=   0.000000000000E+00
    PC002002=   1.000000000000E+00
    PC003002=   0.000000000000E+00
    PC004002=   0.000000000000E+00
    PC001003=   0.000000000000E+00
    PC002003=   0.000000000000E+00
    PC003003=   1.000000000000E+00
    PC004003=   0.000000000000E+00
    PC001004=   0.000000000000E+00
    PC002004=   0.000000000000E+00
    PC003004=   0.000000000000E+00
    PC004004=   1.000000000000E+00
    CTYPE1  = 'RA---SIN'
    CRVAL1  =   2.889721000000E+02
    CDELT1  =  -1.953125000000E-03
    CRPIX1  =   5.130000000000E+02
    CUNIT1  = 'deg     '
    CTYPE2  = 'DEC--SIN'
    CRVAL2  =  -7.466052777778E+01
    CDELT2  =   1.953125000000E-03
    CRPIX2  =   5.130000000000E+02
    CUNIT2  = 'deg     '
    CTYPE3  = 'STOKES  '
    CRVAL3  =   1.000000000000E+00
    CDELT3  =   1.000000000000E+00
    CRPIX3  =   1.000000000000E+00
    CUNIT3  = '        '
    CTYPE4  = 'FREQ    '
    CRVAL4  =   1.445000000000E+09
    CDELT4  =   1.600000000000E+07
    CRPIX4  =   1.000000000000E+00
    CUNIT4  = 'HZ      '
    PV2_1   =   0.000000000000E+00
    PV2_2   =   0.000000000000E+00
    RESTFREQ=   1.445000000000E+09 /Rest Frequency (Hz)
    ALTRVAL =  -0.000000000000E+00 /Alternate frequency reference value
    ALTRPIX =   1.000000000000E+00 /Alternate frequency reference pixel
    VELREF  =                    3 /1 LSR, 2 HEL, 3 OBS, +256 Radio
    COMMENT casacore non-standard usage: 4 LSD, 5 GEO, 6 SOU, 7 GAL
    TELESCOP= 'MeerKAT '
    OBSERVER= 'CASA simulator'
    DATE-OBS= '2012-03-21T00:00:00.000000'
    TIMESYS = 'UTC     '
    OBSRA   =   2.889721000000E+02
    OBSDEC  =  -7.466052777778E+01
    OBSGEO-X=   5.111202828133E+06
    OBSGEO-Y=   2.001309252764E+06
    OBSGEO-Z=  -3.237339358474E+06
    OBJECT  = 'KAT7_1445_1x16_12h'
    TELESCOP= 'MeerKAT '
    INSTRUME= 'MeerKAT '
    DISTANCE=   0.000000000000E+00
    DATE    = '2013-01-15T15:21:51.222000' /Date FITS file was written
    ORIGIN  = 'CASA casacore alma-evla '
    """
    for k in list(header_dict.keys()):
        prihdr.set(k, header_dict[k])
    hdulist.writeto(os.path.join(out_dir, fname))



#!/usr/bin/env python
#    Upload antenna positions from local file to remote telescope
# Copyright (c) Tim Molteno 2017-2021.

import argparse
import json

import numpy as np



from tart.operation import settings
from tart_tools.api_handler import AuthorizedAPIhandler

def main():
    parser = argparse.ArgumentParser(
        description="Upload antenna positions from local file to remote telescope",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--api",
        required=False,
        default="https://tart.elec.ac.nz/signal",
        help="Telescope API server URL.",
    )
    parser.add_argument("--pw", default="password", type=str, help="API password")
    parser.add_argument("--file", type=str, required=True, help="JSON file  of 'antenna_positions' to upload")
    parser.add_argument("--rotate", type=float, default=0.0, help="Rotate the antenna positions (degrees)")

    ARGS = parser.parse_args()

    with open(ARGS.file, "r") as f:
        data = f.read()
    positions_dict = json.loads(data)

    if "antenna_positions" in positions_dict:
        api_dict = {}
        api_dict['antenna_positions'] = positions_dict['antenna_positions']

        if (ARGS.rotate != 0.0):
            original_positions = api_dict['antenna_positions']

            print(f"Rotating positions by {ARGS.rotate} degrees")
            rot_rad = np.radians(ARGS.rotate)

            new_positions = settings.rotate_location(
                np.degrees(rot_rad), np.array(original_positions).T
            )
            pos_list = (np.array(new_positions).T).tolist()
            api_dict["antenna_positions"] = pos_list

        api = AuthorizedAPIhandler(ARGS.api, ARGS.pw)
        print(f"Uploading {api_dict}")
        resp = api.post_payload_with_token(
            "calibration/antenna_positions", api_dict["antenna_positions"])
        print(f"SUCCESS {resp}")
    else:
        raise Exception("JSON file should have an element called 'antenna_positions'")



#!/usr/bin/env python
#
# Get  Data from the TART telescope archive in the cloud (Made by Max Scheel)
#
# Copyright (c) Tim Molteno 2023.
# ~/.tartvenv/bin/pip3 install minio tqdm
#
import os
import datetime

from tqdm import tqdm
from minio import Minio

from tart.util import utc

MINIO_API_HOST = "s3.max.ac.nz"
BUCKET_NAME = "tart-hdf"


def handle_archive_request(target, num_observations, output_dir,
                           start_str, duration_str):
    #
    # Used by the command line utility (tart_get_archive_data), as well as to get calibration data
    #
    if not os.path.exists(output_dir):
        os.makedirs(output_dir, exist_ok=True)

    duration = float(duration_str)
    duration_dt = datetime.timedelta(days=0, hours=0,
                                     minutes=duration, seconds=0)

    if start_str[0] == '-':
        start_offset = float(start_str[1:])
        utc_now = utc.now()
        delta_t = datetime.timedelta(days=0, hours=0,
                                     minutes=start_offset, seconds=0)

        start_datetime = utc_now - delta_t
    else:
        start_datetime = utc.from_string(start_str)

    stop_datetime = start_datetime + duration_dt
    print(f"start: {start_datetime}")
    print(f"stop:  {stop_datetime}")

    # Do some logic to avoid boundaries and get the prefix as long as possible
    # This saves us from downloading the entire year's worth of objects just to
    # get some within the same day...
    prefix = f"{target}/vis"
    if stop_datetime.year == start_datetime.year:
        prefix += f"/{start_datetime.year}"
    if stop_datetime.month == start_datetime.month:
        prefix += f"/{start_datetime.month}"
    if stop_datetime.day == start_datetime.day:
        prefix += f"/{start_datetime.day}"

    print(f"Getting data from path {prefix}")

    client = Minio(MINIO_API_HOST, secure=True)
    objects = client.list_objects(BUCKET_NAME, prefix=prefix, recursive=True,
                                  include_user_meta=True)

    desired_objects = []
    for o in objects:
        # print(o.object_name, o.last_modified)
        if (o.last_modified >= start_datetime) and (o.last_modified <= stop_datetime):
            desired_objects.append(o)

    N = num_observations
    if N > 0:
        steps = max(len(desired_objects) // N, 1)
        data_to_get = list(desired_objects)[::steps]
    else:
        data_to_get = list(desired_objects)

    index = 0
    for item in tqdm(data_to_get):
        # print(item.object_name)
        #
        dirname, fname = os.path.split(item.object_name)
        fname = f"obs_{index:05d}.hdf"
        fname_out = os.path.join(output_dir, fname)
        index = index+1
        client.fget_object(BUCKET_NAME, item.object_name, fname_out)



#!/usr/bin/env python
#
# Get  Data from the TART telescope archive in the cloud (Made by Max Scheel)
#
# Copyright (c) Tim Molteno 2023-2025.
# ~/.tartvenv/bin/pip3 install minio tqdm
#
import argparse

from tart_tools.archive_handler import handle_archive_request


def main():
    parser = argparse.ArgumentParser(description='Get data from the TART s3 bucket',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('--dir', required=False, default=".",
                        help="Output directory")

    parser.add_argument('--name', required=False,
                        default=None,
                        help="Output file name prefix - eg 'out_' would produce files 'out_1.hdf', out_2.hdf ...")

    parser.add_argument('--target', required=False,
                        default='signal', help="Telescope name in s3 bucket.")

    parser.add_argument('--n', required=False, type=int,
                        default=-1, help="Number of HDF vis files. (-1 means all)")

    parser.add_argument('--start', required=False, default="-60",
                        help="Start time (negative means offset from now).")
    parser.add_argument('--duration', required=False,  default="0",
                        help="Number of minutes to sample for")

    ARGS = parser.parse_args()

    handle_archive_request(target=ARGS.target,
                           num_observations=ARGS.n,
                           output_dir=ARGS.dir,
                           start_str=ARGS.start,
                           duration_str=ARGS.duration)



#!/usr/bin/env python
#    Upload gains from local file to api.
#    Max Scheel 2017 - max@max.ac.nz
#    Tim Molteno 2017-2021.

import argparse
import json
import logging
import numpy as np


from tart_tools import api_handler

logger = logging.getLogger()


def test_gains(config, ARGS):

    new_gain = np.random.uniform(size=config.get_num_antenna())
    new_phase = np.random.uniform(size=config.get_num_antenna())
    api_dict = {}
    api_dict['gain'] = new_gain.tolist()
    api_dict['phase_offset'] = new_phase.tolist()
    api = api_handler.AuthorizedAPIhandler(ARGS.api, ARGS.pw)
    resp = api_handler.upload_gain(api, api_dict)

    gains = api_handler.download_current_gain(api)

    recovered_gain = np.array(gains['gain'])
    recovered_phase = np.array(gains['phase_offset'])

    test_gsum = np.sum(recovered_gain - new_gain)
    if test_gsum != 0:
        print(f"FAIL sum of differences = {test_gsum}")
        print(f"uploaded:  {new_gain}")
        print(f"recovered: {recovered_gain}")
        return

    test_gsum = np.sum(recovered_phase - new_phase)
    if test_gsum != 0:
        print(f"PHASE FAIL sum of differences = {test_gsum}")
        print(f"uploaded:  {new_phase}")
        print(f"recovered: {recovered_phase}")
        return

    print("SUCCESS: gains")


def test_antpos(config, ARGS):

    new_ant_pos = np.random.uniform(size=(config.get_num_antenna(),3))

    api_dict = {}
    api_dict['antenna_positions'] = new_ant_pos.tolist()

    api = api_handler.AuthorizedAPIhandler(ARGS.api, ARGS.pw)

    resp = api.post_payload_with_token(
        "calibration/antenna_positions", api_dict['antenna_positions'])

    ant_pos = api.get("imaging/antenna_positions")
    print("Recovered Antenna Positions")
    print(ant_pos)

    recovered_ant_pos = np.array(ant_pos)

    test_gsum = np.sum(recovered_ant_pos - new_ant_pos)
    if test_gsum != 0:
        print(f"FAIL sum of differences = {test_gsum}")
        print(f"uploaded:  {new_ant_pos}")
        print(f"recovered: {recovered_ant_pos}")
    else:
        print("SUCCESS: antpos")


def main():
    parser = argparse.ArgumentParser(
        description="Test a telescope API by setting stuff and checking it",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--api",
        required=False,
        default="https://api.elec.ac.nz/tart/mu-udm",
        help="Telescope API server URL."
    )
    parser.add_argument("--pw", default="password",
                        type=str, help="API password")

    logger.setLevel(logging.DEBUG)

    ARGS = parser.parse_args()

    api = api_handler.APIhandler(ARGS.api)
    config = api_handler.get_config(api)

    test_gains(config, ARGS)
    test_antpos(config, ARGS)




#!/usr/bin/env python
#
#    Download Antenna Positions
#    Tim Molteno 2017 - tim@elec.ac.nz

import argparse
import json

from tart_tools.api_handler import APIhandler

def main():
    parser = argparse.ArgumentParser(
        description="Save antenna positions from api to local file",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--api",
        required=False,
        default="https://tart.elec.ac.nz/signal",
        help="Telescope API server URL.",
    )
    parser.add_argument(
        "--file",
        default="antenna_positions.json",
        type=str,
        help="local file to dump antenna positions",
    )

    ARGS = parser.parse_args()

    api = APIhandler(ARGS.api)

    pos = api.get("imaging/antenna_positions")

    out_json = {}
    out_json["antenna_positions"] = pos
    with open(ARGS.file, "w") as outfile:
        json.dump(out_json, outfile, sort_keys=True, indent=4, ensure_ascii=False)



#!/usr/bin/env python
#
# Change the telescope operating mode using the RESTful the API on a TART radio telescope
# Tim Molteno 2021 tim@elec.ac.nz
#
import argparse

from tart_tools.api_handler import AuthorizedAPIhandler

def main():
    parser = argparse.ArgumentParser(
        description="Change telescope mode",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--api",
        required=False,
        default="https://tart.elec.ac.nz/signal",
        help="Telescope API server URL.",
    )
    parser.add_argument("--pw", default="password", type=str, help="API password")

    mode_group = parser.add_mutually_exclusive_group()
    mode_group.add_argument('--raw', action="store_true", help="Set raw data mode.")
    mode_group.add_argument('--vis', action="store_true", help="Set vis mode")
    mode_group.add_argument("--mode", type=str, required=False, default=None, help="New mode (vis/raw)")

    ARGS = parser.parse_args()

    mode = None
    if ARGS.raw:
        mode = "raw"
    if ARGS.vis:
        mode = "vis"

    if ARGS.mode is not None:
        mode = ARGS.mode

    if mode is None:
        raise RuntimeError("Either --mode, --raw or --vis must be specified")

    api = AuthorizedAPIhandler(ARGS.api, ARGS.pw)
    resp = api.post_payload_with_token(f"mode/{mode}", {})
    print(f"New mode: {resp}")



#!/usr/bin/env python
#    Upload gains from local file to api.
#    Max Scheel 2017 - max@max.ac.nz
#    Tim Molteno 2017-2021.

import argparse
import json

from tart_tools.api_handler import AuthorizedAPIhandler, upload_gain

def main():
    parser = argparse.ArgumentParser(
        description="Upload gains from local file to api",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--api",
        required=False,
        default="https://tart.elec.ac.nz/signal",
        help="Telescope API server URL.",
    )
    parser.add_argument("--pw", default="password", type=str, help="API password")
    parser.add_argument("--gains", type=str, required=True, help="local file to upload")

    ARGS = parser.parse_args()

    with open(ARGS.gains, "r") as f:
        data = f.read()
    gains_dict = json.loads(data)

    api_dict = {}
    api_dict['gain'] = gains_dict['gain']
    api_dict['phase_offset'] = gains_dict['phase_offset']
    api = AuthorizedAPIhandler(ARGS.api, ARGS.pw)
    resp = upload_gain(api, api_dict)
    print("SUCCESS")



#!/usr/bin/env python
#    Download Raw Data from a TART telescope
#    The telescope should be in RAW mode if you want it to capture raw data every minute.
#
# Copyright (c) Tim Molteno 2017-2024.

import argparse
import logging
import os
import urllib.request
import urllib.error
import urllib.parse
import time

from tart_tools.api_handler import APIhandler, download_file, sha256_checksum

logger = logging.getLogger()


def main():
    parser = argparse.ArgumentParser(
        description="Download data from the telescope",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--api",
        required=False,
        default="https://tart.elec.ac.nz/signal",
        help="Telescope API server URL.",
    )
    parser.add_argument(
        "--pw", default="password", required=False, type=str, help="API password"
    )
    parser.add_argument(
        "--dir", type=str, default=".", help="local directory to download"
    )
    parser.add_argument(
        "--n", type=int, default=-1, help="Stop after downloading this many files."
    )
    parser.add_argument(
        "--raw", action="store_true", help="Download Raw Data in HDF format"
    )
    parser.add_argument(
        "--vis", action="store_true", help="Download Visibility Data in HDF format"
    )
    parser.add_argument(
        "--file", type=str, default=None, help="Set the name of the output file"
    )

    ARGS = parser.parse_args()

    logger.setLevel(logging.DEBUG)
    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    # create formatter
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    # add formatter to ch
    ch.setFormatter(formatter)
    # add ch to logger
    logger.addHandler(ch)

    #api = AuthorizedAPIhandler(ARGS.api, ARGS.pw)
    api = APIhandler(ARGS.api)

    logger.info(f"Downloading Data from {ARGS.api}")
    os.makedirs(ARGS.dir, exist_ok=True)

    tart_endpoint = ARGS.api + "/"

    if not (ARGS.raw or ARGS.vis):
        raise RuntimeError("Either --raw or --vis must be specified")

    if (ARGS.n > 1) and (ARGS.file is not None):
        raise RuntimeError("Cannot specify both --n > 1 and --file")

    while True:
        resp_vis = []
        resp_raw = []
        if ARGS.vis:
            resp_vis = api.get("vis/data")
        if ARGS.raw:
            resp_raw = api.get("raw/data")

        try:
            for entry in (resp_raw + resp_vis)[0:ARGS.n]:
                if "filename" in entry:

                    data_url = urllib.parse.urljoin(tart_endpoint, entry["filename"])

                    file_name = data_url.split("/")[-1]
                    if ((ARGS.n == 1) and (ARGS.file is not None)):
                        file_name = ARGS.file

                    file_path = os.path.join(ARGS.dir, file_name)

                    if os.path.isfile(file_path):
                        if sha256_checksum(file_path) == entry["checksum"]:
                            logger.info("Skipping {}".format(file_path))
                        else:
                            logger.info("Corrupted File: {}".format(file_path))
                            os.remove(file_path)
                            download_file(data_url, entry["checksum"], file_path)
                    else:
                        download_file(data_url, entry["checksum"], file_path)

        except Exception as e:
            logger.exception(e)
        finally:
            if ARGS.n > 0:
                break
            logger.info("Pausing.")
            time.sleep(2)



#!/usr/bin/env python
#
# Generate and Image from the RESTful the API on a TART radio telescope
# Copyright (c) Tim Molteno 2017-2022.
#
import matplotlib
import os

if os.name == "posix" and "DISPLAY" not in os.environ:
    matplotlib.use("Agg")
import matplotlib.pyplot as plt

import logging
import argparse
import sys
import threading
import datetime
import json

import numpy as np

from tart_tools import api_imaging
from tart_tools import api_handler
from tart.operation import settings
from tart.imaging import elaz
from tart.util import utc

from copy import deepcopy

logger = logging.getLogger()


def handle_image(args, img, n_bin, title, time_repr, source_json=None):
    """ This function manages the output of an image, drawing sources e.t.c."""
    image_title = "{}_{}".format(title, time_repr)
    if args.fits:
        fname = "{}.fits".format(image_title)
        api_imaging.save_fits_image(
            img, fname=fname, out_dir=args.dir, timestamp=time_repr
        )
        print("Generating {}".format(fname))
    if args.PNG or args.display:
        api_imaging.make_image(plt, img, image_title, n_bin, source_json, args.healpix)
    if args.PNG:
        fname = "{}.png".format(image_title)
        plt.savefig(os.path.join(args.dir, fname))
        print("Generating {}".format(fname))
    if args.display:
        plt.show()


def image_stats(img):
    rsky = np.real(img.flatten())

    npix = rsky.shape[0]

    max_p = np.max(rsky)
    sdev_p = np.std(rsky)
    min_p = np.min(rsky)
    mean_p = np.mean(rsky)
    med_p = np.median(rsky)
    deviation = np.abs(med_p - rsky)
    mad_p = np.median(deviation)

    logger.info(
        "'N_s':{}, 'S/N': {}, 'min': {}, 'max': {}, 'mean': {}, 'sdev': {}, 'R_mad': {}, 'MAD': {}, 'median': {}".format(
            npix,
            (max_p / sdev_p),
            min_p,
            max_p,
            mean_p,
            sdev_p,
            (max_p / mad_p),
            mad_p,
            med_p,
        )
    )

    return max_p, min_p, mad_p


def main():
    PARSER = argparse.ArgumentParser(
        description="Generate an image using the web api ofs a TART radio telescope.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    PARSER.add_argument(
        "--api",
        required=False,
        default="https://tart.elec.ac.nz/signal",
        help="Telescope API server URL.",
    )
    PARSER.add_argument(
        "--catalog",
        required=False,
        default="https://tart.elec.ac.nz/catalog",
        help="Catalog API URL.",
    )
    PARSER.add_argument(
        "--file",
        required=False,
        default=None,
        help="Get data from a calibration JSON file.",
    )

    PARSER.add_argument(
        "--gains",
        required=False,
        default=None,
        help="Use a local JSON file containing antenna gains to create the image image.",
    )
    PARSER.add_argument(
        "--vis",
        required=False,
        default=None,
        help="Use a local JSON file containing the visibilities to create the image image.",
    )
    PARSER.add_argument("--dir", required=False, default=".", help="Output directory.")
    PARSER.add_argument(
        "--rotation",
        type=float,
        default=0.0,
        help="Apply rotation (in degrees) to the antenna positions.",
    )
    PARSER.add_argument(
        "--nfft",
        type=int,
        default=10,
        help="Log(2) of the number of points in the fft.",
    )

    PARSER.add_argument(
        "--dirty", action="store_true", help="Create a direct IFFT dirty image."
    )
    PARSER.add_argument(
        "--difmap",
        action="store_true",
        help="Use difmap to generate a CLEAN image (requires extenal difmap executable).",
    )
    PARSER.add_argument(
        "--aipy", action="store_true", help="Use AIPY to generate a CLEAN image."
    )
    PARSER.add_argument(
        "--moresane",
        action="store_true",
        help="Use MORESANE to generate a wavelet based compressed representation (requires extenal difmap executable).",
    )
    PARSER.add_argument("--beam", action="store_true", help="Generate beam image")

    PARSER.add_argument(
        "--healpix",
        action="store_true",
        help="Use HealPIX to map the resulting image onto a projected circle.",
    )

    PARSER.add_argument(
        "--display", action="store_true", help="Display Image to the user"
    )
    PARSER.add_argument(
        "--log", action="store_true", help="Output the log of the image"
    )
    PARSER.add_argument(
        "--fits", action="store_true", help="Generate a FITS format image"
    )
    PARSER.add_argument(
        "--PNG", action="store_true", help="Generate a PNG format image"
    )
    PARSER.add_argument(
        "--show-sources",
        action="store_true",
        help="Show known sources on images (only works on PNG).",
    )

    source_json = None

    ARGS = PARSER.parse_args()

    logger.setLevel(logging.DEBUG)
    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    # create formatter
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    # add formatter to ch
    ch.setFormatter(formatter)
    # add ch to logger
    logger.addHandler(ch)

    if ARGS.file:
        logger.info("Getting Data from file: {}".format(ARGS.file))
        # Load data from a JSON file
        with open(ARGS.file, "r") as json_file:
            calib_info = json.load(json_file)

        info = calib_info["info"]
        ant_pos = calib_info["ant_pos"]
        config = settings.from_api_json(info["info"], ant_pos)

        flag_list = []  # [4, 5, 14, 22]

        original_positions = deepcopy(config.get_antenna_positions())

        gains_json = calib_info["gains"]
        gains = np.asarray(gains_json["gain"])
        phase_offsets = np.asarray(gains_json["phase_offset"])
        config = settings.from_api_json(info["info"], ant_pos)

        measurements = []
        for d in calib_info["data"]:
            vis_json, source_json = d
            cv, timestamp = api_imaging.vis_calibrated(
                vis_json, config, gains, phase_offsets, flag_list
            )
            src_list = elaz.from_json(source_json, 0.0)

    else:
        logger.info("Getting Data from API: {}".format(ARGS.api))

        api = api_handler.APIhandler(ARGS.api)
        config = api_handler.get_config(api)

        if ARGS.gains is None:
            gains = api.get("calibration/gain")
        else:
            with open(ARGS.gains, "r") as json_file:
                gains = json.load(json_file)

        if ARGS.vis is None:
            vis_json = api.get("imaging/vis")
        else:
            with open(ARGS.vis, "r") as json_file:
                vis_json = json.load(json_file)

        ts = api_imaging.vis_json_timestamp(vis_json)
        if ARGS.show_sources:
            cat_url = api.catalog_url(lon=config.get_lon(),
                                      lat=config.get_lat(),
                                      datestr=utc.to_string(ts))
            source_json = api.get_url(cat_url)

        logger.info("Data Download Complete")

        cv, timestamp = api_imaging.vis_calibrated(
            vis_json, config, gains["gain"], gains["phase_offset"], flag_list=[]
        )

    api_imaging.rotate_vis(
        ARGS.rotation, cv, reference_positions=deepcopy(config.get_antenna_positions())
    )
    n_bin = 2 ** ARGS.nfft

    time_repr = "{:%Y_%m_%d_%H_%M_%S_%Z}".format(timestamp)

    # Image informstion
    # prihdr.set('DATE', timestamp)
    # prihdr.set('DATE-OBS', timestamp)
    #
    # prihdr.set('TIMESYS', 'UTC')
    # prihdr.set('INSTRUME', 'TART')
    # prihdr.set('TELESCOP', 'TART')
    # prihdr.set('OBSERVER' 'CASA simulator')
    # prihdr.set('ORIGIN', 'tart_tools tart.elec.ac.nz ')
    #
    # prihdr.set('OBSRA', 2.889721000000E+02)
    # prihdr.set('OBSDEC', -7.466052777778E+01)
    # prihdr.set('OBSGEO-X', 5.111202828133E+06)
    # prihdr.set('OBSGEO-Y', 2.001309252764E+06)
    # prihdr.set('OBSGEO-Z', -3.237339358474E+06)

    # Processing

    if ARGS.dirty or ARGS.moresane or ARGS.aipy:
        cal_ift, cal_extent, n_fft, bin_width = api_imaging.image_from_calibrated_vis(
            cv, nw=n_bin / 4, num_bin=n_bin
        )

        ## Scale to both abs and MAD
        img = np.abs(cal_ift)
        max_p, min_p, mad_p = image_stats(img)
        ift_scaled = (img - min_p) / mad_p

    if ARGS.beam or ARGS.moresane or ARGS.aipy:
        beam = np.abs(
            api_imaging.beam_from_calibrated_vis(cv, nw=n_bin / 4, num_bin=n_bin)
        )

    if ARGS.difmap:
        fits_bin = 2 ** 12

        fitsfile = "{}.uvfits".format(time_repr)
        uvfitspath = os.path.join(ARGS.dir, fitsfile)
        fits_gen = api_imaging.get_uv_fits(cv)
        fits_gen.write(uvfitspath)

        commands = """
! Basic imaging instructions by Tim Molteno, based around the final_clean script
! by Dan Homan,  dhoman@nrao.edu
debug = False
observe {}
mapsize {}, {}
docont=false
mapcolor color
integer clean_niter;
float clean_gain; clean_gain = 0.03
float dynam;
float flux_peak;

! Define the inner loop as a macro.

float flux_cutoff
float dyn_range
float last_in_rms
float in_rms
float target_rms
float V_rms

#+map_residual \
flux_peak = peak(flux);\
flux_cutoff = imstat(rms) * dynam;\
while(abs(flux_peak)>flux_cutoff);\
 clean clean_niter,clean_gain;\
 flux_cutoff = imstat(rms) * dynam;\
 flux_peak = peak(flux);\
end while

! The following macro stops
! when the the in_frame RMS matches the
! the V RMS OR if there is not improvement
! in the in_frame RMS

#+deep_map_residual \
in_rms = imstat(rms);\
print "Target RMS: ", target_rms, "  In Frame RMS: ", in_rms;\
while(in_rms > target_rms);\
 clean min(200*(in_rms/target_rms),1000),clean_gain;\
 last_in_rms = in_rms;\
 in_rms = imstat(rms);\
 print "Target RMS: ", target_rms, "  In Frame RMS: ", in_rms;\
 if(last_in_rms <= in_rms);\
  in_rms = target_rms;\
 end if;\
 selfcal;\
end while


! select V and get the V_rms for comparison
select v
clrmod true,true,true
delwin
uvw 0,-2
uvtaper 0
V_rms = imstat(rms)

! select the stokes to clean
select i

! clear previous model
clrmod true,true,true

! delete any windows
delwin

! remove any tapering
uvtaper 0

print "*********** FIRST TRY SUPER-UNIFORM WEIGHTING **********"
print "**** -- only if dynamic range is higher than 10 -- *****"

dynam = 10
clean_niter = 50
uvw 20,-1
map_residual
uvw 10,-1
map_residual
clean_niter = 50

print "*********** REGULAR UNIFORM WEIGHTING NEXT ***************"

uvw 2,-1
dynam = 6
map_residual
print "********** DEEP CLEANING AT NATURAL WEIGHTING **************"
uvw 0,-2
! now let clean go deep
target_rms = imstat(noise)/8
if(target_rms < imstat(rms))
  deep_map_residual
else
  ! clean 1 component just to have something to restore
  clean 1, clean_gain
end if

in_rms = imstat(rms)
print "********** FINAL CLEAN IS FINISHED **************"
print "Target RMS was: ", target_rms, " Reached RMS: ", in_rms
print "For comparison uncleaned V RMS is: ", V_rms
print "*************************************************"

device {}_difmap.png/png
mappl cln
!wmap {}_clean.fits
exit
""".format(
            uvfitspath, fits_bin, 1.3 * 648000000.0 / fits_bin, uvfitspath, uvfitspath
        )
        f = open("difmap_cmds", "w")
        f.write(commands)
        f.close()
        os.system("/usr/local/bin/difmap < difmap_cmds")
        # os.system("vips scale {}_clean.fits {}_grey.png".format(uvfitsfile, uvfitsfile))

    if ARGS.moresane:
        from pymoresane.main import DataImage

        degrees_per_pixel = 180.0 / len(beam)

        sane = DataImage(
            img / np.sum(beam),
            beam,
            mask_data=None,
            cdelt1=degrees_per_pixel,
            cdelt2=degrees_per_pixel,
        )
        logger = sane.make_logger("INFO")

        beam_width_degrees = 3.0
        stop_scale = ARGS.nfft - 1  # Maximum size feature (full image width)
        start_scale = int(np.log2(n_bin * beam_width_degrees / 180.0))

        print("Start Scale {}".format(start_scale))
        sane.moresane(
            subregion=None,
            sigma_level=4.0,
            loop_gain=0.05,
            tolerance=0.75,
            accuracy=1e-6,
            major_loop_miter=100,
            minor_loop_miter=50,
            all_on_gpu=False,
            decom_mode="ser",
            core_count=1,
            conv_device="cpu",
            conv_mode="circular",
            extraction_mode="cpu",
            enforce_positivity=True,
            edge_suppression=True,
            edge_offset=int(n_bin / 30),
            flux_threshold=0,
            neg_comp=False,
            edge_excl=0,
            int_excl=0,
        )
        sane.restore()

    if ARGS.aipy:
        import aipy

        clean, info = aipy.deconv.clean(
            img,
            beam,
            gain=0.1,
            maxiter=1000,
            tol=1e-5,
            verbose=True,
            stop_if_div=True,
            pos_def=False,
        )

        if True:
            b_restore = np.fft.fftshift(beam)
        else:
            # b_restore = a.img.gaussian_beam(3, shape=img.shape) #* np.max(beam)
            b_center = len(beam) / 2
            b_width_px = int(3.0 * n_bin / 180)
            b_restore_shifted = np.zeros_like(beam)

            y, x = np.ogrid[-b_center:b_center, -b_center:b_center]
            mask = x * x + y * y <= b_width_px ** 2

            b_restore_shifted[mask] = np.fft.fftshift(beam)[mask]
            b_restore = np.fft.fftshift(b_restore_shifted)

        restored_image = np.abs(
            np.fft.ifft2(np.fft.fft2(clean) * np.fft.fft2(b_restore))
        )  # Cropped PSF Restoration

        print("Clean", info["success"], np.min(restored_image), np.max(restored_image))

    # Do output images

    if ARGS.moresane:
        handle_image(ARGS, sane.restored, n_bin, "MORESANE", time_repr, source_json)

    if ARGS.beam:
        if ARGS.log:
            beam = np.log10(beam)
        handle_image(ARGS, beam, n_bin, "beam", time_repr)

    if ARGS.dirty:
        handle_image(ARGS, ift_scaled, n_bin, "dirty", time_repr, source_json)

    if ARGS.aipy:
        handle_image(ARGS, restored_image, n_bin, "clean", time_repr, source_json)



import unittest
import logging
import numpy as np
from copy import deepcopy

from tart_tools import api_imaging
from tart_tools import api_handler
from tart.util import utc
from tart.util import angle

from tart.imaging import antenna_model
from tart.imaging import radio_source
from tart.imaging import correlator
from tart.imaging import location
from tart.imaging import calibration
from tart.imaging import elaz
from tart.simulation import skymodel
from tart.simulation import antennas
from tart.simulation import radio


logger = logging.getLogger()


class TestApi(unittest.TestCase):

    def setUp(self):
        self.api = 'https://api.elec.ac.nz/tart/nz-dunedin'
        self.pw = 'sharkbait'

        api = api_handler.APIhandler(self.api)

        self.config = api_handler.get_config(api)
        self.ant_pos = self.config.get_antenna_positions()

        gains = api.get("calibration/gain")

        vis_json = api.get("imaging/vis")

        ts = api_imaging.vis_json_timestamp(vis_json)

        cat_url = api.catalog_url(lon=self.config.get_lon(),
                                  lat=self.config.get_lat(),
                                  datestr=utc.to_string(ts))
        self.source_json = api.get_url(cat_url)

        logger.info("Data Download Complete")

        self.cv, self.timestamp = api_imaging.vis_calibrated(
            vis_json, self.config,
            gains["gain"],
            gains["phase_offset"],
            flag_list=[]
        )

    def get_fake_vis(self):

        loc = location.Location(angle.from_dms(self.config.get_lat()),
                                angle.from_dms(self.config.get_lon()),
                                self.config.get_alt())

        ANTS = [antennas.Antenna(loc, pos) for pos in self.ant_pos]
        num_ant = len(ANTS)
        ANT_MODELS = [antenna_model.GpsPatchAntenna() for i in range(num_ant)]
        NOISE_LVLS = np.zeros(num_ant)
        RAD = radio.Max2769B(n_samples=2**12, noise_level=NOISE_LVLS)

        sim_sky = skymodel.Skymodel(0, location=loc, gps=0, thesun=0, known_cosmic=0)

        timestamp = utc.now()

        # ############## HOUR HAND ###########################
        #
        # The pattern rotates once every 12 hours
        #
        hour_hand = timestamp.hour*30.0 + timestamp.minute/2.0

        self.hour_sources = [{'el': el, 'az': -hour_hand} for el in [85, 75, 65, 55]]
        for m in self.hour_sources:
            sim_sky.add_src(radio_source.ArtificialSource(loc, timestamp, r=1e6, el=m['el'], az=m['az']))

        # ############## MINUTE HAND ###########################
        #
        # The pattern rotates once every 1 hour
        #
        minute_hand = timestamp.minute*6.0 + timestamp.second/10.0

        self.minute_sources = [{'el': el, 'az': -minute_hand} for el in [90, 80, 70, 60, 50, 40, 30]]
        for m in self.minute_sources:
            sim_sky.add_src(radio_source.ArtificialSource(loc, timestamp, r=1e6, el=m['el'], az=m['az']))

        sources = sim_sky.gen_photons_per_src(timestamp, radio=RAD,
                                              config=self.config, n_samp=1)

        v_sim = antennas.antennas_simp_vis(
            ANTS, ANT_MODELS, sources, timestamp, self.config, RAD.noise_level
        )

        cv = calibration.CalibratedVisibility(v_sim)

        return cv

    def test_imaging(self):

        cv = self.get_fake_vis()
        rotation = 0.0
        api_imaging.rotate_vis(
            rotation, cv, reference_positions=deepcopy(self.ant_pos)
        )
        n_bin = 2 ** 8

        cal_ift, cal_extent, n_fft, bin_width = api_imaging.image_from_calibrated_vis(
            cv, nw=n_bin / 4, num_bin=n_bin
        )

        img = np.abs(cal_ift)
        max_p = np.max(img)
        min_p = np.min(img)
        mad_p = np.std(img)
        ift_scaled = (img - min_p) / mad_p

        import matplotlib.pyplot as plt

        for m in self.minute_sources:
            src = elaz.ElAz(m['el'], m['az'])
            x, y = src.get_px(n_bin)
            print(x, y)
            ift_scaled[x, y] = -1
        plt.imshow(ift_scaled)

        plt.show()

        for m in self.minute_sources:
            src = elaz.ElAz(m['el'], m['az'])
            x, y = src.get_px(n_bin)
            self.assertGreater(img[y, x], max_p/3)



"""
  TART API handler class
  Tim Molteno 2017-2023 - tim@elec.ac.nz
  Max Scheel 2017 - max@max.ac.nz
"""

import hashlib
import json
import logging
import os
import shutil
import urllib.error
import urllib.parse
import urllib.request
import tempfile

import requests
from requests.adapters import HTTPAdapter
from tart.operation import settings
from tart.util import utc
from urllib3.util.retry import Retry

# Default timeout
TIMEOUT = 15.0

logger = logging.getLogger()


def sha256_checksum(filename, block_size=65536):
    sha256 = hashlib.sha256()
    with open(filename, "rb") as f:
        for block in iter(lambda: f.read(block_size), b""):
            sha256.update(block)
    return sha256.hexdigest()


def download_file(url, checksum=0, file_path=None):
    logger.info(f"Download_file({url}, {checksum}) -> {file_path}")

    # Download the file from `url` to a temporary file and save it locally under `file_path`:
    with urllib.request.urlopen(url) as response, tempfile.NamedTemporaryFile(delete=False) as temp_file:
        shutil.copyfileobj(response, temp_file)
    shutil.move(temp_file.name, file_path)

    if checksum:
        downloaded_checksum = sha256_checksum(file_path)
        if downloaded_checksum != checksum:
            logger.info(
                f"Removing file: Checksum failed\n{checksum}\n{downloaded_checksum}"
            )
            os.remove(file_path)


class APIhandler:
    def __init__(self, api_root):
        self.root = api_root
        self.token = None

    def url(self, path):
        return f"{self.root}/api/v1/{path}"

    def catalog_url(
        self,
        lon,  # degrees
        lat,
        catalog="https://tart.elec.ac.nz/catalog",
        datestr=None
    ):
        if datestr is None:
            datestr = utc.to_string(utc.now())
        return f"{catalog}/catalog?lat={lat}&lon={lon}&date={datestr}"

    def get(self, path):
        return self.get_url(self.url(path))

    def get_url(self, url):

        s = requests.Session()

        retries = Retry(total=5,
                        backoff_factor=1,
                        status_forcelist=[429, 500, 502, 503, 504])

        s.mount('https://', HTTPAdapter(max_retries=retries))

        r = s.get(url, timeout=TIMEOUT)
        r.raise_for_status()
        return json.loads(r.text)


class AuthorizedAPIhandler(APIhandler):
    """
    This object allows an authorized API call.

    Example:

        from tart_tools.api_handler import AuthorizedAPIhandler

        if __name__ == '__main__':
            parser = argparse.ArgumentParser(description='Change telescope mode',
                            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
            parser.add_argument('--api', required=False, default='https://tart.elec.ac.nz/signal',
                            help="Telescope API server URL.")
            parser.add_argument('--pw', default='password', type=str, help='API password')
            parser.add_argument('--mode', type=str, required=True, help='New mode (vis/raw)')

            ARGS = parser.parse_args()

            api = AuthorizedAPIhandler(ARGS.api, ARGS.pw)

            resp = api.post_payload_with_token('mode', ARGS.mode)


    """

    def __init__(self, api_root, pw):
        APIhandler.__init__(self, api_root=api_root)
        self.pw = pw
        self.token = None
        self.refresh_token = None
        self.login()

    def login(self):
        payload_dict = {"username": "admin", "password": self.pw}
        resp_json = self.__post_payload("/auth", payload_dict=payload_dict)
        if "access_token" in resp_json:
            self.token = resp_json["access_token"]
            self.refresh_token = resp_json["refresh_token"]
        else:
            raise RuntimeError("Authorization failed. Wrong pw?")

    def refresh_access_token(self):
        r = requests.post(
            self.url("refresh"), headers=self.__get_refresh_header(), timeout=TIMEOUT
        )
        resp_json = json.loads(r.text)
        if "access_token" in resp_json:
            self.token = resp_json["access_token"]
            logger.info("refreshed token")

    def __post_payload(self, path, payload_dict):
        """ Currently only used for login() """
        r = requests.post(self.url(path), json=payload_dict, timeout=TIMEOUT)
        r.raise_for_status()
        return json.loads(r.text)

    def __get_header(self):
        if self.token is None:
            raise RuntimeError("login required")
        return {"Authorization": "JWT " + self.token}

    def __get_refresh_header(self):
        if self.refresh_token is None:
            raise RuntimeError("login required")
        return {"Authorization": "JWT " + self.refresh_token}

    # TODO: catch the requests result that corresponds to a failed login (authorization expired)
    # and re-authorize automatically.
    def put(self, path, **kwargs):
        r = requests.put(
            self.url(path), headers=self.__get_header(), timeout=TIMEOUT, **kwargs
        )
        if r.status_code in (requests.codes.unauthorized, requests.codes.internal_server_error):
            self.refresh_access_token()
            return self.put(path, **kwargs)
        logger.info(f"put response {r}")
        ret = json.loads(r.text)
        r.raise_for_status()
        return ret

    def post(self, path, **kwargs):
        r = requests.post(
            self.url(path), headers=self.__get_header(), timeout=TIMEOUT, **kwargs
        )
        if r.status_code in (requests.codes.unauthorized, requests.codes.internal_server_error):
            self.refresh_access_token()
            return self.post(path, **kwargs)
        logger.info(f"post response {r}")
        ret = json.loads(r.text)
        r.raise_for_status()
        return ret

        # #logger.info(r)
        # ret = json.loads(r.text)
        # if "status" in ret and "sub_status" in ret:
        #     if (ret["status"] == 401) and (ret["sub_status"] == 101):
        #         self.refresh_access_token()
        #         return self.post(path, **kwargs)
        # r.raise_for_status()
        # return ret

    def post_with_token(self, path):
        return self.post(path)

    def post_payload_with_token(self, path, payload_dict):
        return self.post(path, json=payload_dict)


def set_mode(api, mode):
    resp_json = api.post_with_token("mode/" + mode)
    return resp_json


def download_current_gain(api):
    gains = api.get("calibration/gain")
    return gains


def upload_gain(api, gain_dict):
    resp = api.post_payload_with_token("calibration/gain", gain_dict)
    logger.info("SUCCESS")
    return resp


def get_config(api):
    info = api.get("info")
    ant_pos = api.get("imaging/antenna_positions")
    return settings.from_api_json(info["info"], ant_pos)
